import pytest
import time
import tempfile
from pathlib import Path

from unoserver import server


@pytest.fixture(scope="session")
def server_fixture():
    with tempfile.TemporaryDirectory() as tmpuserdir:
        user_installation = Path(tmpuserdir).as_uri()
        srvr = server.UnoServer(user_installation=user_installation)
        process = srvr.start()
        # Give libreoffice a chance to start
        time.sleep(8)
        yield process  # provide the fixture value
        print("Teardown Unoserver")
        srvr.stop()



"""Unoserver unit tests"""
import os

from unittest import mock
from unoserver import server

TEST_DOCS = os.path.join(os.path.abspath(os.path.split(__file__)[0]), "documents")


@mock.patch("threading.Thread")
@mock.patch("subprocess.Popen")
def test_server_params(popen_mock, thread_mock):
    srv = server.UnoServer(port="2203", uno_port="2202")
    srv.start()
    popen_mock.assert_called_with(
        [
            "libreoffice",
            "--headless",
            "--invisible",
            "--nocrashreport",
            "--nodefault",
            "--nologo",
            "--nofirststartwizard",
            "--norestore",
            f"-env:UserInstallation={srv.user_installation}",
            "--accept=socket,host=127.0.0.1,port=2202,tcpNoDelay=1;urp;StarOffice.ComponentContext",
        ]
    )


@mock.patch("threading.Thread")
@mock.patch("subprocess.Popen")
def test_server_ipv6_params(popen_mock, thread_mock):
    srv = server.UnoServer(interface="::", port="2203", uno_port="2202")
    srv.start()
    popen_mock.assert_called_with(
        [
            "libreoffice",
            "--headless",
            "--invisible",
            "--nocrashreport",
            "--nodefault",
            "--nologo",
            "--nofirststartwizard",
            "--norestore",
            f"-env:UserInstallation={srv.user_installation}",
            "--accept=socket,host=127.0.0.1,port=2202,tcpNoDelay=1;urp;StarOffice.ComponentContext",
        ]
    )



"""Tests that start a real unoserver and does real things"""

import io
import os
import pytest
import re
import socket
import subprocess
import sys
import tempfile
import time

from xmlrpc.client import Fault
from unoserver import client


TEST_DOCS = os.path.join(os.path.abspath(os.path.split(__file__)[0]), "documents")


@pytest.mark.parametrize("filename", ["simple.odt", "simple.xlsx"])
def test_pdf_conversion(server_fixture, filename):
    infile = os.path.join(TEST_DOCS, filename)

    with tempfile.NamedTemporaryFile(suffix=".pdf") as outfile:
        # Let Libreoffice write to the file and close it.
        sys.argv = ["unoconverter", infile, outfile.name]
        client.converter_main()

        # We now open it to check it, we can't use the outfile object,
        # it won't reflect the external changes.
        with open(outfile.name, "rb") as testfile:
            start = testfile.readline()
            assert start.startswith(b"%PDF-1.")


class FakeStdio(io.BytesIO):
    """A BytesIO with a buffer attribute, usable to send binary stdin data"""

    @property
    def buffer(self):
        return self


@pytest.mark.parametrize("filename", ["simple.odt", "simple.xlsx"])
def test_stdin_stdout(server_fixture, monkeypatch, filename):
    with open(os.path.join(TEST_DOCS, filename), "rb") as infile:
        infile_stream = FakeStdio(infile.read())

    outfile_stream = FakeStdio()

    monkeypatch.setattr("sys.stdin", infile_stream)
    monkeypatch.setattr("sys.stdout", outfile_stream)

    sys.argv = ["unoconverter", "-", "-", "--convert-to", "pdf"]
    client.converter_main()

    outfile_stream.seek(0)
    start = outfile_stream.readline()
    assert start.startswith(b"%PDF-1.")


def test_csv_conversion(server_fixture):
    conv = client.UnoClient()
    infile = os.path.join(TEST_DOCS, "simple.xlsx")

    with tempfile.NamedTemporaryFile(suffix=".csv") as outfile:
        # Let Libreoffice write to the file and close it.
        conv.convert(inpath=infile, outpath=outfile.name)
        # We now open it to check it, we can't use the outfile object,
        # it won't reflect the external changes.
        with open(outfile.name, "rb") as testfile:
            contents = testfile.readline()
            assert contents == b"This,Is,A,Simple,Excel,File\n"
            contents = testfile.readline()
            assert contents == b"1,2,3,4,5,6\n"


def test_impossible_conversion(server_fixture):
    conv = client.UnoClient()
    infile = os.path.join(TEST_DOCS, "simple.odt")

    with tempfile.NamedTemporaryFile(suffix=".xls") as outfile:
        # Let Libreoffice write to the file and close it.
        with pytest.raises(Fault) as e:
            conv.convert(inpath=infile, outpath=outfile.name)
            assert "Could not find an export filter" in e


def test_multiple_servers(server_fixture):
    # The server fixture should already have started a server.
    # Make sure we can start a second one.
    cmd = ["unoserver", "--uno-port=2102", "--port=2103"]
    process = subprocess.Popen(cmd)
    try:
        # Wait for it to start
        time.sleep(5)
        # Make sure the process is still running, meaning return_code is None
        assert process.returncode is None

        # Make a conversion
        conv = client.UnoClient(port="2103")
        infile = os.path.join(TEST_DOCS, "simple.odt")
        with tempfile.NamedTemporaryFile(suffix=".pdf") as outfile:
            conv.convert(inpath=infile, outpath=outfile.name)

    finally:
        # Now kill the process
        process.terminate()
        # Wait for it to terminate
        process.wait(30)
        # And verify that it was killed
        assert process.returncode == 0


def test_unknown_outfile_type(server_fixture):
    infile = os.path.join(TEST_DOCS, "simple.odt")

    with tempfile.NamedTemporaryFile(suffix=".bog") as outfile:
        sys.argv = ["unoconverter", infile, outfile.name]
        # Type detection should fail, as it's not a .doc file:
        with pytest.raises(Fault) as e:
            client.converter_main()
        assert "Unknown export file type" in e.value.faultString


@pytest.mark.parametrize("filename", ["simple.odt", "simple.xlsx"])
def test_explicit_export_filter(server_fixture, filename):
    infile = os.path.join(TEST_DOCS, filename)

    # We use an extension that's not .pdf to verify that the converter does not auto-detect filter based on extension
    with tempfile.NamedTemporaryFile(suffix=".csv") as outfile:
        sys.argv = [
            "unoconverter",
            "--filter",
            "writer_pdf_Export",
            infile,
            outfile.name,
        ]
        client.converter_main()

        # We now open it to check it, we can't use the outfile object,
        # it won't reflect the external changes.
        with open(outfile.name, "rb") as testfile:
            start = testfile.readline()
            assert start.startswith(b"%PDF-1.")


@pytest.mark.parametrize("filename", ["simple.odt", "simple.xlsx"])
def test_invalid_explicit_export_filter_prints_available_filters(
    caplog, server_fixture, filename
):
    infile = os.path.join(TEST_DOCS, filename)

    # We use an extension that's not .pdf to verify that the converter does not auto-detect filter based on extension
    with tempfile.NamedTemporaryFile(suffix=".csv") as outfile:
        sys.argv = ["unoconverter", "--filter", "asdasdasd", infile, outfile.name]
        try:
            client.converter_main()
        except RuntimeError:
            errstr = caplog.text
            print(errstr)
            print("=" * 30)
            assert "Office Open XML Text" in errstr
            assert "writer8" in errstr
            assert "writer_pdf_Export" in errstr


def test_update_index(server_fixture):
    infile = os.path.join(TEST_DOCS, "index-with-fields.odt")

    with tempfile.NamedTemporaryFile(suffix=".rtf") as outfile:
        # Let Libreoffice write to the file and close it.
        sys.argv = ["unoconverter", infile, outfile.name]
        client.converter_main()

        # We now open it to check it, we can't use the outfile object,
        # it won't reflect the external changes.
        with open(outfile.name, "rb") as testfile:
            # The timestamp in Header 2 should appear exactly twice after update
            matches = re.findall(b"13:18:27", testfile.read())
            assert len(matches) == 2

        with tempfile.NamedTemporaryFile(suffix=".rtf") as outfile:
            # Let Libreoffice write to the file and close it.
            sys.argv = ["unoconverter", "--dont-update-index", infile, outfile.name]
            client.converter_main()

            with open(outfile.name, "rb") as testfile:
                # The timestamp in Header 2 should appear exactly once
                matches = re.findall(b"13:18:27", testfile.read())
                assert len(matches) == 1


def test_convert_not_local():
    hostname = socket.gethostname()
    cmd = ["unoserver", "--uno-port=2104", "--port=2105", f"--interface={hostname}"]
    process = subprocess.Popen(cmd)
    try:
        # Wait for it to start
        time.sleep(5)
        # Make sure the process is still running, meaning return_code is None
        assert process.returncode is None

        # Make a conversion
        infile = os.path.join(TEST_DOCS, "simple.odt")
        with tempfile.NamedTemporaryFile(suffix=".pdf") as outfile:
            sys.argv = [
                "unoconverter",
                "--host",
                hostname,
                "--port=2105",
                infile,
                outfile.name,
            ]
            client.converter_main()

            with open(outfile.name, "rb") as testfile:
                start = testfile.readline()
                assert start.startswith(b"%PDF-1.")

    finally:
        # Now kill the process
        process.terminate()
        # Wait for it to terminate
        process.wait(30)
        # And verify that it was killed
        assert process.returncode == 0


# This currently does not work on Ubuntu 20.04.
def skip_test_compare_not_local():
    hostname = socket.gethostname()
    cmd = ["unoserver", "--uno-port=2104", "--port=2105", f"--interface={hostname}"]
    process = subprocess.Popen(cmd)
    try:
        # Wait for it to start
        time.sleep(5)
        # Make sure the process is still running, meaning return_code is None
        assert process.returncode is None

        # Make a comparison
        infile1 = os.path.join(TEST_DOCS, "simple.odt")
        infile2 = os.path.join(TEST_DOCS, "index-with-fields.odt")
        with tempfile.NamedTemporaryFile(suffix=".pdf") as outfile:
            sys.argv = [
                "unoconverter",
                "--host",
                hostname,
                "--port=2105",
                "--input-filter=odt",
                infile1,
                infile2,
                outfile.name,
            ]
            client.comparer_main()

            with open(outfile.name, "rb") as testfile:
                start = testfile.readline()
                assert start.startswith(b"%PDF-1.")

    finally:
        # Now kill the process
        process.terminate()
        # Wait for it to terminate
        process.wait(30)
        # And verify that it was killed
        assert process.returncode == 0



from __future__ import annotations

import argparse
import logging
import os
import shutil
import signal
import socket
import subprocess
import sys
import tempfile
import threading
import time
import platform
import xmlrpc.server
from importlib import metadata
from pathlib import Path

from concurrent import futures

from unoserver import converter, comparer

API_VERSION = "3"
__version__ = metadata.version("unoserver")
logger = logging.getLogger("unoserver")


class XMLRPCServer(xmlrpc.server.SimpleXMLRPCServer):
    def __init__(
        self,
        addr: tuple[str, int],
        allow_none: bool = False,
    ) -> None:
        addr_info = socket.getaddrinfo(addr[0], addr[1], proto=socket.IPPROTO_TCP)

        if len(addr_info) == 0:
            raise RuntimeError(
                f"Could not get interface information for {addr[0]}:{addr[1]}"
            )

        self.address_family = addr_info[0][0]
        self.socket_type = addr_info[0][1]
        super().__init__(addr=addr_info[0][4], allow_none=allow_none)


class UnoServer:
    def __init__(
        self,
        interface="127.0.0.1",
        port="2003",
        uno_interface="127.0.0.1",
        uno_port="2002",
        user_installation=None,
        conversion_timeout=None,
    ):
        self.interface = interface
        self.uno_interface = uno_interface
        self.port = port
        self.uno_port = uno_port
        self.user_installation = user_installation
        self.conversion_timeout = conversion_timeout
        self.libreoffice_process = None
        self.xmlrcp_thread = None
        self.xmlrcp_server = None
        self.intentional_exit = False

    def start(self, executable="libreoffice"):
        logger.info(f"Starting unoserver {__version__}.")

        connection = (
            "socket,host=%s,port=%s,tcpNoDelay=1;urp;StarOffice.ComponentContext"
            % (self.uno_interface, self.uno_port)
        )

        # I think only --headless and --norestore are needed for
        # command line usage, but let's add everything to be safe.
        cmd = [
            executable,
            "--headless",
            "--invisible",
            "--nocrashreport",
            "--nodefault",
            "--nologo",
            "--nofirststartwizard",
            "--norestore",
            f"-env:UserInstallation={self.user_installation}",
            f"--accept={connection}",
        ]

        logger.info("Command: " + " ".join(cmd))
        self.libreoffice_process = subprocess.Popen(cmd)
        self.xmlrcp_thread = threading.Thread(None, self.serve)

        def signal_handler(signum, frame):
            self.intentional_exit = True
            logger.info("Sending signal to LibreOffice")
            try:
                self.libreoffice_process.send_signal(signum)
            except ProcessLookupError as e:
                # 3 means the process is already dead
                if e.errno != 3:
                    raise

            if self.xmlrcp_server is not None:
                self.stop()  # Ensure the server stops

        signal.signal(signal.SIGTERM, signal_handler)
        signal.signal(signal.SIGINT, signal_handler)

        # Signal SIGHUP is available only in Unix systems
        if platform.system() != "Windows":
            signal.signal(signal.SIGHUP, signal_handler)

        time.sleep(10)

        self.xmlrcp_thread.start()

        # Give the thread time to start
        time.sleep(2)
        # Check if it succeeded
        if not self.xmlrcp_thread.is_alive():
            logger.info("Failed to start servers")
            self.stop()
            return None

        return self.libreoffice_process

    def serve(self):
        # Create server
        with XMLRPCServer((self.interface, int(self.port)), allow_none=True) as server:
            self.conv = converter.UnoConverter(
                interface=self.uno_interface, port=self.uno_port
            )
            self.comp = comparer.UnoComparer(
                interface=self.uno_interface, port=self.uno_port
            )

            self.xmlrcp_server = server
            server.register_introspection_functions()

            @server.register_function
            def info():
                import_filters = self.conv.get_filter_names(
                    self.conv.get_available_import_filters()
                )
                export_filters = self.conv.get_filter_names(
                    self.conv.get_available_export_filters()
                )
                return {
                    "unoserver": __version__,
                    "api": API_VERSION,
                    "import_filters": import_filters,
                    "export_filters": export_filters,
                }

            @server.register_function
            def convert(
                inpath=None,
                indata=None,
                outpath=None,
                convert_to=None,
                filtername=None,
                filter_options=[],
                update_index=True,
                infiltername=None,
            ):
                if indata is not None:
                    indata = indata.data

                with futures.ThreadPoolExecutor() as executor:
                    future = executor.submit(
                        self.conv.convert,
                        inpath,
                        indata,
                        outpath,
                        convert_to,
                        filtername,
                        filter_options,
                        update_index,
                        infiltername,
                    )
                    try:
                        return future.result(timeout=self.conversion_timeout)
                    except futures.TimeoutError:
                        logger.error(
                            "Conversion timeout, terminating conversion and exiting."
                        )
                        self.conv.local_context.dispose()
                        self.libreoffice_process.terminate()
                        raise

            @server.register_function
            def compare(
                oldpath=None,
                olddata=None,
                newpath=None,
                newdata=None,
                outpath=None,
                filetype=None,
            ):
                if olddata is not None:
                    olddata = olddata.data
                if newdata is not None:
                    newdata = newdata.data

                with futures.ThreadPoolExecutor() as executor:
                    future = executor.submit(
                        self.comp.compare,
                        oldpath,
                        olddata,
                        newpath,
                        newdata,
                        outpath,
                        filetype,
                    )
                try:
                    return future.result(timeout=self.conversion_timeout)
                except futures.TimeoutError:
                    logger.error(
                        "Comparison timeout, terminating conversion and exiting."
                    )
                    self.conv.local_context.dispose()
                    self.libreoffice_process.terminate()
                    raise

            server.serve_forever()

    def stop(self):

        if self.xmlrcp_server is not None:
            self.xmlrcp_server.shutdown()
            # Make a dummy connection to unblock accept() - otherwise it will
            # hang indefinitely in the accept() call.
            # noinspection PyBroadException
            try:
                with socket.create_connection(
                    (self.interface, int(self.port)), timeout=1
                ):
                    pass
            except Exception:
                pass  # Ignore any except

        if self.xmlrcp_thread is not None:
            self.xmlrcp_thread.join()

        if self.libreoffice_process and self.libreoffice_process.poll() is not None:
            self.libreoffice_process.terminate()
            try:
                self.libreoffice_process.wait(10)
            except subprocess.TimeoutExpired:
                logger.info("Signalling harder...")
                self.libreoffice_process.terminate()


def main():
    logging.basicConfig()
    logger.setLevel(logging.INFO)

    parser = argparse.ArgumentParser("unoserver")
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        help="Display version and exit.",
        version=f"{parser.prog} {__version__}",
    )
    parser.add_argument(
        "--interface",
        default="127.0.0.1",
        help="The interface used by the XMLRPC server",
    )
    parser.add_argument(
        "--uno-interface",
        default="127.0.0.1",
        help="The interface used by the Libreoffice UNO server",
    )
    parser.add_argument(
        "--port", default="2003", help="The port used by the XMLRPC server"
    )
    parser.add_argument(
        "--uno-port", default="2002", help="The port used by the Libreoffice UNO server"
    )
    parser.add_argument("--daemon", action="store_true", help="Deamonize the server")
    parser.add_argument(
        "--executable",
        default=None,
        help="The path to the LibreOffice executable, defaults to looking in the path",
    )
    parser.add_argument(
        "--user-installation",
        default=None,
        help="The path to the LibreOffice user profile",
    )
    parser.add_argument(
        "--libreoffice-pid-file",
        "-p",
        default=None,
        help="If set, unoserver will write the Libreoffice PID to this file. If started "
        "in daemon mode, the file will not be deleted when unoserver exits.",
    )
    parser.add_argument(
        "--conversion-timeout",
        type=int,
        help="Terminate Libreoffice and exit if a conversion does not complete in the "
        "given time (in seconds).",
    )
    args = parser.parse_args()

    if args.daemon:
        cmd = sys.argv
        cmd.remove("--daemon")
        proc = subprocess.Popen(cmd)
        return proc.pid

    with tempfile.TemporaryDirectory() as tmpuserdir:
        user_installation = Path(tmpuserdir).as_uri()

        if args.user_installation is not None:
            user_installation = Path(args.user_installation).as_uri()

        if args.uno_port == args.port:
            raise RuntimeError("--port and --uno-port must be different")

        server = UnoServer(
            args.interface,
            args.port,
            args.uno_interface,
            args.uno_port,
            user_installation,
            args.conversion_timeout,
        )

        if args.executable is not None:
            executable = args.executable
        else:
            # Find the executable automatically. I had problems with
            # LibreOffice using 100% if started with the libreoffice
            # executable, so by default try soffice first. Also throwing
            # ooffice in there as a fallback, I don't think it's used any
            # more, but it doesn't hurt to have it there.
            for name in ("soffice", "libreoffice", "ooffice"):
                if (executable := shutil.which(name)) is not None:
                    break

        # If it's daemonized, this returns the process.
        # It returns 0 of getting killed in a normal way.
        # Otherwise it returns 1 after the process exits.
        process = server.start(executable=executable)
        if process is None:
            return 2
        pid = process.pid

        logger.info(f"Server PID: {pid}")

        if args.libreoffice_pid_file:
            with open(args.libreoffice_pid_file, "wt") as upf:
                upf.write(f"{pid}")

        process.wait()

        if not server.intentional_exit:
            logger.error(f"Looks like LibreOffice died. PID: {pid}")

        # The RPC thread needs to be stopped before the process can exit
        server.stop()
        if args.libreoffice_pid_file:
            # Remove the PID file
            os.unlink(args.libreoffice_pid_file)

        try:
            # Make sure it's really dead
            os.kill(pid, 0)

            if server.intentional_exit:
                return 0
            else:
                return 1
        except OSError as e:
            if e.errno == 3:
                # All good, it was already dead.
                return 0
            raise


if __name__ == "__main__":
    main()



import argparse
import logging
import os
import sys
import time

from importlib import metadata
from xmlrpc.client import ServerProxy

__version__ = metadata.version("unoserver")
logger = logging.getLogger("unoserver")

API_VERSION = "3"
SFX_FILTER_IMPORT = 1
SFX_FILTER_EXPORT = 2
DOC_TYPES = {
    "com.sun.star.sheet.SpreadsheetDocument",
    "com.sun.star.text.TextDocument",
    "com.sun.star.presentation.PresentationDocument",
    "com.sun.star.drawing.DrawingDocument",
    "com.sun.star.sdb.DocumentDataSource",
    "com.sun.star.formula.FormulaProperties",
    "com.sun.star.script.BasicIDE",
    "com.sun.star.text.WebDocument",  # Supposedly deprecated? But still around.
}


class UnoClient:
    """An RPC client for Unoserver"""

    def __init__(self, server="127.0.0.1", port="2003", host_location="auto"):
        self.server = server
        self.port = port
        if host_location == "auto":
            if server in ("127.0.0.1", "localhost"):
                self.remote = False
            else:
                self.remote = True
        elif host_location == "remote":
            self.remote = True
        elif host_location == "local":
            self.remote = False
        else:
            raise RuntimeError("host_location can be 'auto', 'remote', or 'local'")

    def _connect(self, proxy, retries=5, sleep=10):
        """Check the connection to the proxy multiple times

        Returns the info() data (unoserver version + filters)"""

        while retries > 0:
            try:
                info = proxy.info()
                if not info["api"] == API_VERSION:
                    raise RuntimeError(
                        f"API Version mismatch. Client {__version__} uses API {API_VERSION} "
                        f"while Server {info['unoserver']} uses API {info['api']}."
                    )
                return info
            except ConnectionError as e:
                logger.debug(f"Error {e.strerror}, waiting...")
                retries -= 1
                if retries > 0:
                    time.sleep(sleep)
                    logger.debug("Retrying...")
                else:
                    raise

    def convert(
        self,
        inpath=None,
        indata=None,
        outpath=None,
        convert_to=None,
        filtername=None,
        filter_options=[],
        update_index=True,
        infiltername=None,
    ):
        """Converts a file from one type to another

        inpath: A path (on the local hard disk) to a file to be converted.

        indata: A byte string containing the file content to be converted.

        outpath: A path (on the local hard disk) to store the result, or None, in which case
                 the content of the converted file will be returned as a byte string.

        convert_to: The extension of the desired file type, ie "pdf", "xlsx", etc.

        filtername: The name of the export filter to use for conversion. If None, it is auto-detected.

        update_index: Updates the index before conversion
        """
        if inpath is None and indata is None:
            raise RuntimeError("Nothing to convert.")

        if inpath is not None and indata is not None:
            raise RuntimeError("You can only pass in inpath or indata, not both.")

        if convert_to is None:
            if outpath is None:
                raise RuntimeError(
                    "If you don't specify an output path, you must specify a file-type."
                )
            else:
                convert_to = os.path.splitext(outpath)[-1].strip(os.path.extsep)

        if inpath:
            if self.remote:
                with open(inpath, "rb") as infile:
                    indata = infile.read()
                    inpath = None
            else:
                inpath = os.path.abspath(inpath)

        if outpath:
            outpath = os.path.abspath(outpath)
            if os.path.isdir(outpath):
                raise ValueError("The outpath can not be a directory")

        with ServerProxy(f"http://{self.server}:{self.port}", allow_none=True) as proxy:
            logger.info("Connecting.")
            logger.debug(f"Host: {self.server} Port: {self.port}")
            info = self._connect(proxy)

            if infiltername and infiltername not in info["import_filters"]:
                existing = "\n".join(sorted(info["import_filters"]))
                logger.critical(
                    f"Unknown import filter: {infiltername}. Available filters:\n{existing}"
                )
                raise RuntimeError("Invalid parameter")

            if filtername and filtername not in info["export_filters"]:
                existing = "\n".join(sorted(info["export_filters"]))
                logger.critical(
                    f"Unknown export filter: {filtername}. Available filters:\n{existing}"
                )
                raise RuntimeError("Invalid parameter")

            logger.info("Converting.")
            result = proxy.convert(
                inpath,
                indata,
                None if self.remote else outpath,
                convert_to,
                filtername,
                filter_options,
                update_index,
                infiltername,
            )
            if result is not None:
                # We got the file back over xmlrpc:
                if outpath:
                    logger.info(f"Writing to {outpath}.")
                    with open(outpath, "wb") as outfile:
                        outfile.write(result.data)
                else:
                    # Return the result as a blob
                    logger.info(f"Returning {len(result.data)} bytes.")
                    return result.data
            else:
                logger.info(f"Saved to {outpath}.")

    def compare(
        self,
        oldpath=None,
        olddata=None,
        newpath=None,
        newdata=None,
        outpath=None,
        filetype=None,
    ):
        """Compare two files and convert the result from one type to another.

        newpath: A path (on the local hard disk) to a file to be compared.

        newdata: A byte string containing the file content to be compared.

        oldpath: A path (on the local hard disk) to another file to be compared.

        olddata: A byte string containing the other file content to be compared.

        outpath: A path (on the local hard disk) to store the result, or None, in which case
                 the content of the converted file will be returned as a byte string.

        convert_to: The extension of the desired file type, ie "pdf", "xlsx", etc.
        """
        if (newpath is None and newdata is None) or (
            oldpath is None and olddata is None
        ):
            raise RuntimeError(
                "Nothing to be compared. You mast pass in newpath or newdata and oldpath or olddata."
            )

        if newpath is not None and newdata is not None:
            raise RuntimeError("You can only pass in newpath or newdata, not both.")

        if oldpath is not None and olddata is not None:
            raise RuntimeError("You can only pass in oldpath or olddata, not both.")

        if outpath is None and filetype is None:
            raise RuntimeError(
                "If you don't specify an resulting filepath, you must specify a file-type."
            )
        elif filetype is None:
            filetype = os.path.splitext(outpath)[-1].strip(os.path.extsep)

        if self.remote:
            if oldpath:
                with open(oldpath, "rb") as infile:
                    olddata = infile.read()
                    oldpath = None

            if newpath:
                with open(newpath, "rb") as infile:
                    newdata = infile.read()
                    newpath = None

        if oldpath:
            oldpath = os.path.abspath(oldpath)
        if newpath:
            newpath = os.path.abspath(newpath)

        with ServerProxy(f"http://{self.server}:{self.port}", allow_none=True) as proxy:
            logger.info("Connecting.")
            logger.debug(f"Host: {self.server} Port: {self.port}")
            self._connect(proxy)

            logger.info("Comparing.")
            result = proxy.compare(
                oldpath,
                olddata,
                newpath,
                newdata,
                outpath if self.server in ("127.0.0.1", "localhost") else None,
                filetype,
            )
            if result is not None:
                # We got the file back over xmlrpc:
                if outpath:
                    logger.info(f"Writing to {outpath}.")
                    with open(outpath, "wb") as outfile:
                        outfile.write(result.data)
                else:
                    # Return the result as a blob
                    logger.info(f"Returning {len(result.data)} bytes.")
                    return result.data
            else:
                logger.info(f"Saved to {outpath}.")


def converter_main():
    logging.basicConfig()

    parser = argparse.ArgumentParser("unoconvert")
    parser.add_argument(
        "infile", help="The path to the file to be converted (use - for stdin)"
    )
    parser.add_argument(
        "outfile", help="The path to the converted file (use - for stdout)"
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        help="Display version and exit.",
        version=f"{parser.prog} {__version__}",
    )
    parser.add_argument(
        "--convert-to",
        help="The file type/extension of the output file (ex pdf). Required when using stdout",
    )
    parser.add_argument(
        "--input-filter",
        help="The LibreOffice input filter to use (ex 'writer8'), if autodetect fails",
    )
    parser.add_argument(
        "--output-filter",
        "--filter",
        default=None,
        help="The export filter to use when converting. It is selected automatically if not specified.",
    )
    parser.add_argument(
        "--filter-options",
        "--filter-option",
        default=[],
        action="append",
        help="Pass an option for the export filter, in name=value format, or for positional parameters, "
        "a comma separated list. Use true/false for boolean values. Can be repeated for multiple options.",
    )
    parser.add_argument(
        "--update-index",
        action="store_true",
        help="Updates the indexes before conversion. Can be time consuming.",
    )
    parser.add_argument(
        "--dont-update-index",
        action="store_false",
        dest="update_index",
        help="Skip updating the indexes.",
    )
    parser.set_defaults(update_index=True)
    parser.add_argument(
        "--host", default="127.0.0.1", help="The host the server runs on"
    )
    parser.add_argument("--port", default="2003", help="The port used by the server")
    parser.add_argument(
        "--host-location",
        default="auto",
        choices=["auto", "remote", "local"],
        help="The host location determines the handling of files. If you run the client on the "
        "same machine as the server, it can be set to local, and the files are sent as paths. "
        "If they are different machines, it is remote and the files are sent as binary data. "
        "Default is auto, and it will send the file as a path if the host is 127.0.0.1 or "
        "localhost, and binary data for other hosts.",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        dest="verbose",
        help="Increase informational output to stderr.",
    )
    parser.add_argument(
        "--quiet",
        action="store_true",
        dest="quiet",
        help="Decrease informational output to stderr.",
    )
    args = parser.parse_args()

    if args.verbose:
        logger.setLevel(logging.DEBUG)
    elif args.quiet:
        logger.setLevel(logging.CRITICAL)
    else:
        logger.setLevel(logging.INFO)
    if args.verbose and args.quiet:
        logger.debug("Make up your mind, yo!")

    client = UnoClient(args.host, args.port, args.host_location)

    if args.outfile == "-":
        # Set outfile to None, to get the data returned from the function,
        # instead of written to a file.
        args.outfile = None

    if args.infile == "-":
        # Get data from stdin
        indata = sys.stdin.buffer.read()
        args.infile = None
    else:
        indata = None

    result = client.convert(
        inpath=args.infile,
        indata=indata,
        outpath=args.outfile,
        convert_to=args.convert_to,
        filtername=args.output_filter,
        filter_options=args.filter_options,
        update_index=args.update_index,
        infiltername=args.input_filter,
    )

    if args.outfile is None:
        sys.stdout.buffer.write(result)


def comparer_main():
    logging.basicConfig()
    logger.setLevel(logging.INFO)

    parser = argparse.ArgumentParser("unocompare")
    parser.add_argument(
        "oldfile",
        help="The path to the original file to be compared with the modified one (use - for stdin)",
    )
    parser.add_argument(
        "newfile",
        help="The path to the modified file to be compared with the original one (use - for stdin)",
    )
    parser.add_argument(
        "outfile",
        help="The path to the result of the comparison and converted file (use - for stdout)",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        help="Display version and exit.",
        version=f"{parser.prog} {__version__}",
    )
    parser.add_argument(
        "--file-type",
        help="The file type/extension of the result file (ex pdf). Required when using stdout",
    )
    parser.add_argument(
        "--host", default="127.0.0.1", help="The host the server run on"
    )
    parser.add_argument("--port", default="2003", help="The port used by the server")
    parser.add_argument(
        "--host-location",
        default="auto",
        choices=["auto", "remote", "local"],
        help="The host location determines the handling of files. If you run the client on the "
        "same machine as the server, it can be set to local, and the files are sent as paths. "
        "If they are different machines, it is remote and the files are sent as binary data. "
        "Default is auto, and it will send the file as a path if the host is 127.0.0.1 or "
        "localhost, and binary data for other hosts.",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        dest="verbose",
        help="Increase informational output to stderr.",
    )
    parser.add_argument(
        "--quiet",
        action="store_true",
        dest="quiet",
        help="Decrease informational output to stderr.",
    )
    args = parser.parse_args()

    if args.verbose:
        logger.setLevel(logging.DEBUG)
    elif args.quiet:
        logger.setLevel(logging.CRITICAL)
    else:
        logger.setLevel(logging.INFO)
    if args.verbose and args.quiet:
        logger.debug("Make up your mind, yo!")

    client = UnoClient(args.host, args.port, args.host_location)

    if args.outfile == "-":
        # Set outfile to None, to get the data returned from the function,
        # instead of written to a file.
        args.outfile = None

    if args.oldfile == "-" and args.newfile == "-":
        raise RuntimeError("You can't read both files from stdin")

    if args.oldfile == "-":
        # Get data from stdin
        olddata = sys.stdin.buffer.read()
        newdata = None
        args.oldfile = None

    elif args.newfile == "-":
        newdata = sys.stdin.buffer.read()
        olddata = None
        args.newfile = None
    else:
        olddata = newdata = None

    result = client.compare(
        oldpath=args.oldfile,
        olddata=olddata,
        newpath=args.newfile,
        newdata=newdata,
        outpath=args.outfile,
        filetype=args.file_type,
    )

    if args.outfile is None:
        sys.stdout.buffer.write(result)



try:
    import uno
except ImportError:
    raise ImportError(
        "Could not find the 'uno' library. This package must be installed with a Python "
        "installation that has a 'uno' library. This typically means you should install "
        "it with the same Python executable as your Libreoffice installation uses."
    )

import io
import logging
import os
import unohelper

from pathlib import Path
from com.sun.star.beans import PropertyValue
from com.sun.star.io import XOutputStream

logger = logging.getLogger("unoserver")

SFX_FILTER_IMPORT = 1
SFX_FILTER_EXPORT = 2
DOC_TYPES = {
    "com.sun.star.sheet.SpreadsheetDocument",
    "com.sun.star.text.TextDocument",
    "com.sun.star.presentation.PresentationDocument",
    "com.sun.star.drawing.DrawingDocument",
    "com.sun.star.sdb.DocumentDataSource",
    "com.sun.star.formula.FormulaProperties",
    "com.sun.star.script.BasicIDE",
    "com.sun.star.text.WebDocument",  # Supposedly deprecated? But still around.
}


def prop2dict(properties):
    return {p.Name: p.Value for p in properties}


def get_doc_type(doc):
    for t in DOC_TYPES:
        if doc.supportsService(t):
            return t

    # LibreOffice opened it, but it's not one of the known document types.
    # This really should only happen if a future version of LibreOffice starts
    # adding document types, which seems unlikely.
    raise RuntimeError(
        "The input document is of an unknown document type. This is probably a bug.\n"
        "Please create an issue at https://github.com/unoconv/unoserver."
    )


class OutputStream(unohelper.Base, XOutputStream):
    def __init__(self):
        self.buffer = io.BytesIO()

    def closeOutput(self):
        pass

    def writeBytes(self, seq):
        self.buffer.write(seq.value)


class UnoConverter:
    """The class that performs the conversion

    Don't use this directly, instead use the client.UnoConverter.
    """

    def __init__(self, interface="127.0.0.1", port="2002"):
        logger.info("Starting UnoConverter.")

        self.local_context = uno.getComponentContext()
        self.resolver = self.local_context.ServiceManager.createInstanceWithContext(
            "com.sun.star.bridge.UnoUrlResolver", self.local_context
        )
        self.context = self.resolver.resolve(
            f"uno:socket,host={interface},port={port};urp;StarOffice.ComponentContext"
        )
        self.service = self.context.ServiceManager
        self.desktop = self.service.createInstanceWithContext(
            "com.sun.star.frame.Desktop", self.context
        )
        self.filter_service = self.service.createInstanceWithContext(
            "com.sun.star.document.FilterFactory", self.context
        )
        self.type_service = self.service.createInstanceWithContext(
            "com.sun.star.document.TypeDetection", self.context
        )
        self._export_filters = None
        self._import_filters = None

    def find_filter(self, import_type, export_type):
        for export_filter in self.get_available_export_filters():
            if export_filter["DocumentService"] != import_type:
                continue
            if export_filter["Type"] != export_type:
                continue

            # There is only one possible filter per import and export type,
            # so the first one we find is correct
            return export_filter["Name"]

        # No filter found
        return None

    def get_available_import_filters(self):
        # Doing this call for some reason uses up memory each time, so we do it
        # only once, and cache it here:
        if self._import_filters is not None:
            return self._import_filters

        # List import filters. You can only search on module, iflags and eflags,
        # so the import and export types we have to test in a loop
        import_filters = self.filter_service.createSubSetEnumerationByQuery(
            "getSortedFilterList():iflags=1"
        )

        self._import_filters = []
        while import_filters.hasMoreElements():
            self._import_filters.append(prop2dict(import_filters.nextElement()))

        return self._import_filters

    def get_available_export_filters(self):
        # Doing this call for some reason uses up memory each time, so we do it
        # only once, and cache it here:
        if self._export_filters is not None:
            return self._export_filters

        # List export filters. You can only search on module, iflags and eflags,
        # so the import and export types we have to test in a loop
        export_filters = self.filter_service.createSubSetEnumerationByQuery(
            "getSortedFilterList():iflags=2"
        )

        self._export_filters = []
        while export_filters.hasMoreElements():
            self._export_filters.append(prop2dict(export_filters.nextElement()))

        return self._export_filters

    def get_filter_names(self, filters):
        names = {}
        for flt in filters:
            # Add all names and exstensions, etc in a mapping to the internal
            # Libreoffice name, so we can map it.
            # The actual name:
            names[flt["Name"]] = flt["Name"]
            # UserData sometimes has file extensions, etc.
            # Skip empty data, and those weird file paths, and "true"...
            for name in filter(
                lambda x: x and x != "true" and "." not in x, flt["UserData"]
            ):
                names[name] = flt["Name"]
        return names

    def convert(
        self,
        inpath=None,
        indata=None,
        outpath=None,
        convert_to=None,
        filtername=None,
        filter_options=[],
        update_index=True,
        infiltername=None,
    ):
        """Converts a file from one type to another

        inpath: A path (on the local hard disk) to a file to be converted.

        indata: A byte string containing the file content to be converted.

        outpath: A path (on the local hard disk) to store the result, or None, in which case
                 the content of the converted file will be returned as a byte string.

        convert_to: The extension of the desired file type, ie "pdf", "xlsx", etc.

        filtername: The name of the export filter to use for conversion. If None, it is auto-detected.

        filter_options: A list of output filter options as strings, in a "OptionName=Value" format.

        update_index: Updates the index before conversion

        infiltername: The name of the input filter, ie "writer8", "PowerPoint 3", etc.

        You must specify the inpath or the indata, and you must specify and outpath or a convert_to.
        """
        input_props = (PropertyValue(Name="ReadOnly", Value=True),)
        if infiltername:
            infilters = self.get_filter_names(self.get_available_import_filters())
            if infiltername in infilters:
                input_props += (
                    PropertyValue(Name="FilterName", Value=infilters[infiltername]),
                )
            else:
                raise ValueError(
                    f"There is no '{infiltername}' import filter. Available filters: {sorted(infilters.keys())}"
                )

        if inpath:
            # TODO: Verify that inpath exists and is openable, and that outdir exists, because uno's
            # exceptions are completely useless!

            if not Path(inpath).exists():
                raise RuntimeError(f"Path {inpath} does not exist.")

            # Load the document
            logger.info(f"Opening {inpath} for input")
            import_path = uno.systemPathToFileUrl(os.path.abspath(inpath))

        elif indata:
            # The document content is passed in as a byte string
            logger.info("Opening private:stream for input")
            old_stream = self.service.createInstanceWithContext(
                "com.sun.star.io.SequenceInputStream", self.context
            )
            old_stream.initialize((uno.ByteSequence(indata),))
            input_props += (PropertyValue(Name="InputStream", Value=old_stream),)
            import_path = "private:stream"

        document = self.desktop.loadComponentFromURL(
            import_path, "_default", 0, input_props
        )

        if document is None:
            # Could not load document, fail
            if not inpath:
                inpath = "<remote file>"
            if not infiltername:
                infiltername = "default"

            error = f"Could not load document {inpath} using the {infiltername} filter."
            logger.error(error)
            raise RuntimeError(error)

        if update_index:
            # Update document indexes
            for ii in range(2):
                # At first, update Table-of-Contents.
                # ToC grows, so page numbers grow too.
                # On second turn, update page numbers in ToC.
                try:
                    document.refresh()
                    indexes = document.getDocumentIndexes()
                except AttributeError:
                    # The document doesn't implement the XRefreshable and/or
                    # XDocumentIndexesSupplier interfaces
                    break
                else:
                    for i in range(0, indexes.getCount()):
                        indexes.getByIndex(i).update()

        # Now do the conversion
        try:
            # Figure out document type:
            import_type = get_doc_type(document)

            if outpath:
                export_path = uno.systemPathToFileUrl(os.path.abspath(outpath))
            else:
                export_path = "private:stream"

            # Figure out the output type:
            if convert_to:
                export_type = self.type_service.queryTypeByURL(
                    f"file:///dummy.{convert_to}"
                )
            else:
                export_type = self.type_service.queryTypeByURL(export_path)

            if not export_type:
                if convert_to:
                    extension = convert_to
                else:
                    extension = os.path.splitext(outpath)[-1]
                raise RuntimeError(
                    f"Unknown export file type, unknown extension '{extension}'"
                )

            if filtername is not None:
                available_filter_names = self.get_filter_names(
                    self.get_available_export_filters()
                )
                if filtername not in available_filter_names:
                    raise RuntimeError(
                        f"There is no '{filtername}' export-filter. Available filters: {sorted(available_filter_names)}"
                    )
            else:
                filtername = self.find_filter(import_type, export_type)
                if filtername is None:
                    raise RuntimeError(
                        f"Could not find an export filter from {import_type} to {export_type}"
                    )

            logger.info(f"Exporting to {outpath}")
            logger.info(
                f"Using {filtername} export filter from {infiltername} to {export_type}"
            )

            export_filter_data = []
            export_filter_options = []

            for option in filter_options:
                if "=" in option:
                    option_name, option_value = option.split("=", maxsplit=1)
                else:
                    option_name = None
                    option_value = option

                if option_value == "false":
                    option_value = False
                elif option_value == "true":
                    option_value = True
                elif option_value.isdecimal():
                    option_value = int(option_value)

                if option_name is not None:
                    export_filter_data.append(PropertyValue(Name=option_name, Value=option_value))
                else:
                    export_filter_options.append(PropertyValue(Name="FilterOptions", Value=option_value))

            output_props = (
                PropertyValue(Name="FilterName", Value=filtername),
                PropertyValue(Name="Overwrite", Value=True),
            )
            if outpath is None:
                output_stream = OutputStream()
                output_props += (
                    PropertyValue(Name="OutputStream", Value=output_stream),
                )
            if export_filter_data:
                output_props += (
                    PropertyValue(
                        Name="FilterData",
                        Value=uno.Any(
                            "[]com.sun.star.beans.PropertyValue", tuple(export_filter_data)
                        ),
                    ),
                )
            if export_filter_options:
                output_props += tuple(export_filter_options)

            document.storeToURL(export_path, output_props)

        finally:
            document.close(True)

        if outpath is None:
            return output_stream.buffer.getvalue()
        else:
            return None



try:
    import uno
except ImportError:
    raise ImportError(
        "Could not find the 'uno' library. This package must be installed with a Python "
        "installation that has a 'uno' library. This typically means you should install "
        "it with the same Python executable as your Libreoffice installation uses."
    )

import io
import logging
import os
import unohelper

from com.sun.star.beans import PropertyValue
from com.sun.star.io import XOutputStream

logger = logging.getLogger("unoserver")

SFX_FILTER_IMPORT = 1
SFX_FILTER_EXPORT = 2
DOC_TYPES = {
    "com.sun.star.text.TextDocument",  # Only support comparing for writer
}


def prop2dict(properties):
    return {p.Name: p.Value for p in properties}


def get_doc_type(doc):
    for t in DOC_TYPES:
        if doc.supportsService(t):
            return t

    # LibreOffice opened it, but it's not one of the supported document types.
    # This really should only happen if a future version of LibreOffice starts
    # adding document types, which seems unlikely.
    raise RuntimeError(
        "The input document is an unsupported document type for comparing.\n"
        "Please create an issue at https://github.com/unoconv/unoserver."
    )


class OutputStream(unohelper.Base, XOutputStream):
    def __init__(self):
        self.buffer = io.BytesIO()

    def closeOutput(self):
        pass

    def writeBytes(self, seq):
        self.buffer.write(seq.value)


class UnoComparer:
    """The class that performs the comparison

    Don't use this directly, instead use the client.UnoComparer.
    """

    def __init__(self, interface="127.0.0.1", port="2002"):
        logger.info("Starting UnoComparer.")

        self.local_context = uno.getComponentContext()
        self.resolver = self.local_context.ServiceManager.createInstanceWithContext(
            "com.sun.star.bridge.UnoUrlResolver", self.local_context
        )
        self.context = self.resolver.resolve(
            f"uno:socket,host={interface},port={port};urp;StarOffice.ComponentContext"
        )
        self.service = self.context.ServiceManager
        self.desktop = self.service.createInstanceWithContext(
            "com.sun.star.frame.Desktop", self.context
        )
        self.filter_service = self.service.createInstanceWithContext(
            "com.sun.star.document.FilterFactory", self.context
        )
        self.type_service = self.service.createInstanceWithContext(
            "com.sun.star.document.TypeDetection", self.context
        )

    def is_comparable(self, import_type, importOrg_type):
        # List export filters. You can only search on module, iflags and eflags,
        # so the import and export types we have to test in a loop
        export_filters = self.filter_service.createSubSetEnumerationByQuery(
            "getSortedFilterList():iflags=2"
        )

        while export_filters.hasMoreElements():
            # Filter by Filtername here
            export_filter = prop2dict(export_filters.nextElement())
            if export_filter["Type"] != importOrg_type:
                continue
            if export_filter["DocumentService"] != import_type:
                continue

            # There is only one possible DocumentService per import and export type,
            # so the first one we find is correct
            return True

        # No DocumentService found
        return False

    def find_filter(self, import_type, export_type):
        # List export filters. You can only search on module, iflags and eflags,
        # so the import and export types we have to test in a loop
        export_filters = self.filter_service.createSubSetEnumerationByQuery(
            "getSortedFilterList():iflags=2"
        )

        while export_filters.hasMoreElements():
            # Filter DocumentService here
            export_filter = prop2dict(export_filters.nextElement())
            if export_filter["DocumentService"] != import_type:
                continue
            if export_filter["Type"] != export_type:
                continue

            # There is only one possible filter per import and export type,
            # so the first one we find is correct
            return export_filter["Name"]

        # No filter found
        return None

    def compare(
        self,
        oldpath=None,
        olddata=None,
        newpath=None,
        newdata=None,
        outpath=None,
        filetype=None,
    ):
        """Compare two files and convert the result from one type to another.

        inpath: A path (on the local hard disk) to a file to be compared.
        inOrgpath: A path (on the local hard disk) to another file to be compared.

        indata: A byte string containing the file content to be compared.

        outpath: A path (on the local hard disk) to store the result, or None, in which case
                 the content of the converted file will be returned as a byte string.

        filetype: The extension of the desired file type, ie "pdf", "xlsx", etc.
        """
        new_props = (PropertyValue(Name="Hidden", Value=True),)

        if newpath:
            # TODO: Verify that inpath exists and is openable, and that outdir exists, because uno's
            # exceptions are completely useless!

            # Load the document
            logger.info(f"Opening file {newpath}")
            newpath = uno.systemPathToFileUrl(os.path.abspath(newpath))
            # This returned None if the file was locked, I'm hoping the ReadOnly flag avoids that.

        elif newdata:
            # The document content is passed in as a byte string
            new_stream = self.service.createInstanceWithContext(
                "com.sun.star.io.SequenceInputStream", self.context
            )
            new_stream.initialize((uno.ByteSequence(newdata),))
            new_props += (PropertyValue(Name="InputStream", Value=new_stream),)
            newpath = "private:stream"

        new_document = self.desktop.loadComponentFromURL(
            newpath, "_blank", 0, new_props
        )
        new_type = get_doc_type(new_document)

        old_props = (PropertyValue(Name="Hidden", Value=True),)

        if oldpath:
            # TODO: Verify that inpath exists and is openable, and that outdir exists, because uno's
            # exceptions are completely useless!

            # Load the document
            logger.info(f"Opening file {oldpath}")
            oldpath = uno.systemPathToFileUrl(os.path.abspath(oldpath))
            old_props += (PropertyValue(Name="URL", Value=oldpath),)
            # This returned None if the file was locked, I'm hoping the ReadOnly flag avoids that.
            old_type = self.type_service.queryTypeByURL(oldpath)

        elif olddata:
            # The document content is passed in as a byte string
            old_stream = self.service.createInstanceWithContext(
                "com.sun.star.io.SequenceInputStream", self.context
            )
            old_stream.initialize((uno.ByteSequence(newdata),))
            old_props += (PropertyValue(Name="InputStream", Value=new_stream),)
            old_props += (PropertyValue(Name="URL", Value="private:stream"),)
            old_type = self.type_service.queryTypeByDescriptor(old_props, False)[0]

        old_props += (PropertyValue(Name="NoAcceptDialog", Value=True),)

        logger.info(f"Opening original file {oldpath}")

        # Now do the comparison, then the conversion
        try:
            # Figure out document type of import file:
            # Figure out document type of original import file:
            # check that the two type is same
            isComparable = self.is_comparable(new_type, old_type)

            if not isComparable:
                raise RuntimeError("Cannot compare two different type of document!")

            dispatch_helper = self.service.createInstanceWithContext(
                "com.sun.star.frame.DispatchHelper", self.context
            )
            dispatch_helper.executeDispatch(
                new_document.getCurrentController().getFrame(),
                ".uno:CompareDocuments",
                "",
                0,
                old_props,
            )

            if outpath:
                export_path = uno.systemPathToFileUrl(os.path.abspath(outpath))
            else:
                export_path = "private:stream"

            # Figure out the output type:
            if filetype:
                export_type = self.type_service.queryTypeByURL(
                    f"file:///dummy.{filetype}"
                )
            else:
                export_type = self.type_service.queryTypeByURL(export_path)

            if not export_type:
                if filetype:
                    extension = filetype
                else:
                    extension = os.path.splitext(outpath)[-1]
                raise RuntimeError(
                    f"Unknown export file type, unknown extension '{extension}'"
                )

            filtername = self.find_filter(new_type, export_type)
            if filtername is None:
                raise RuntimeError(
                    f"Could not find an export filter from {new_type} to {export_type}"
                )

            logger.info(f"Exporting to {outpath}")
            logger.info(
                f"Using {filtername} export filter from {new_type} to {export_type}"
            )

            output_props = (
                PropertyValue(Name="FilterName", Value=filtername),
                PropertyValue(Name="Overwrite", Value=True),
            )
            if outpath is None:
                output_stream = OutputStream()
                output_props += (
                    PropertyValue(Name="OutputStream", Value=output_stream),
                )
            new_document.storeToURL(export_path, output_props)
            new_document.dispose()

        finally:
            new_document.close(True)

        if outpath is None:
            return output_stream.buffer.getvalue()
        else:
            return None
